# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Date: 07.10.2016                                                            #
# Author: Ole-Johan Skrede                                                    #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
Class for a multivariate Gauss classifier.
"""

# pylint: disable=bad-indentation
# pylint: disable=too-many-instance-attributes
# pylint: disable=import-error
# pylint: disable=too-many-locals
# pylint: disable=too-many-branches
# pylint: disable=too-many-statements

import math

import numpy as np

class Classification(object):
  """Classification class for a multivariate Gaussian classifier.

  training_mask: Image mask used to determine which pixels are in what class
                 for training.
  test_mask: Image mask used to determine which pixels are in what class
             for testing.
  feature_images: List of images where each pixel have a feature value. This image is
                  masked with training_mask and test_mask, when training and
                  evaluating, respectively.
  posterior_images: List of size num_classes. Each element is a posterior image
                    for the corresponding class. This means that each pixel in
                    a class posterior image is valued according to the
                    probability that this pixels belong to that class.
  prop_classified_image: Image classified (segmented) with the proposed
                         classifier. Argmax of posterior_images.
  prior: Vector of size num_classes. Contains prior probability for each class.
  class_means: Vector of size num_classes. Contains the mean value computed
               from the training pixels for each class.
  class_variances: Vector of size num_classes. Contains the variance value
                   computed from the training pixels for each class.
  confusion_matrix: Contingency table with values used for evaluation.
  """
  def __init__(self):
    self.iid = None
    self.training_mask = None
    self.test_mask = None
    self.feature_images = None
    self.posterior_images = None
    self.prop_classified_image = None
    self.classes = []
    self.num_features = None
    self.num_classes = None
    self.prior = None
    self.class_means = None
    self.class_covariances = None
    self.confusion_matrix = None
    self.eval_measures = {}

  def train(self, training_mask, feature_images, iid):
    """Training method"""
    print("Start training")
    # Gaussian classifiers
    self.iid = iid # Assuming independent features with equal variance across classes
                   # In this case we estimate the (shared) variance using a
                   # pooled estimate (see t-test with equal variance on wikipedia)
    self.training_mask = training_mask  # Masked image that define classes
    self.feature_images = feature_images  # list of images
    # We assume that the class label "0" is background, and not a class.
    self.classes = np.unique(self.training_mask)[1:]
    print("Following classes are defined: {0}".format(self.classes))
    self.num_features = len(self.feature_images)
    self.num_classes = len(self.classes)

    self.prior = np.ones(self.num_classes) / float(self.num_classes)  # Uniform

    self.class_means = np.zeros([self.num_classes, self.num_features])
    self.class_covariances = np.zeros([self.num_classes, self.num_features,
                                       self.num_features])

    if self.iid:
      pooling_numerator = 0
      pooling_denominator = 0
    for c in range(self.num_classes):
      num_class_pixels = np.count_nonzero(self.training_mask == self.classes[c])
      class_features = np.zeros([self.num_features, num_class_pixels])
      for f_ind in range(self.num_features):
        feature_image = self.feature_images[f_ind]
        class_features[f_ind, :] = feature_image[np.nonzero(self.training_mask == self.classes[c])]
        if self.iid:
          pooling_numerator += (num_class_pixels - 1)*np.var(class_features[f_ind, :], ddof=1)
      self.class_means[c, :] = np.mean(class_features, axis=1)
      if self.iid:
        pooling_denominator += self.num_features*(num_class_pixels - 1)
      else:
        self.class_covariances[c, :, :] = np.cov(class_features)

    if self.iid:
      pooled_variance_estimate = pooling_numerator / pooling_denominator
      for c in range(self.num_classes):
        self.class_covariances[c, :, :] = pooled_variance_estimate*np.identity(self.num_features)

    # Check if covariance matrix is singular (or ill-conditioned)
    for c in range(self.num_classes):
      det_cov = np.linalg.det(self.class_covariances[c, :, :])
      if det_cov == 0:
        condition_number = np.linalg.cond(self.class_covariances[c, :, :])
        print('Error: Singular covariance matrix for class {}'.format(c))
        print('Determinant = {}'.format(det_cov))
        print('Condition number = {}'.format(condition_number))
        raise ValueError('Singular covariance matrix')


  def classify(self, feature_images, class_means=None, class_covariances=None,
               classes=None):
    """Classify, assumes training is done."""
    print("Start classifying")

    # If you already have trained, pass class means and variances,
    # but if you have trained in the same session, just set it to the trained
    # values.

    if classes is not None:
      self.classes = classes
      self.num_classes = len(class_means)
    if class_means is not None:
      self.class_means = class_means
    if class_covariances is not None:
      self.class_covariances = class_covariances
    self.num_features = len(feature_images)

    print("Number of features used: {}".format(self.num_features))

    M, N = feature_images[0].shape
    self.prop_classified_image = np.zeros([M, N])
    self.posterior_images = np.zeros([self.num_classes, M, N])
    feature_vector = np.zeros(self.num_features)
    for i in range(M):
      for j in range(N):
        for f_ind in range(self.num_features):
          feature_vector[f_ind] = feature_images[f_ind][i, j]

        disc_values, posteriors = self._discriminant_function(feature_vector)

        self.prop_classified_image[i, j] = self.classes[disc_values.argmax()]
        for c in range(self.num_classes):
          self.posterior_images[c, i, j] = posteriors[c]

  def _discriminant_function(self, feature_vector):
    """Computes discriminant function (one value for each class c)"""
    disc_values = np.zeros(self.num_classes)
    log_posteriors = np.zeros(self.num_classes)
    posteriors = np.zeros(self.num_classes)
    if self.prior is None:
      self.prior = np.ones(len(self.classes)) / len(self.classes)
    for c in range(self.num_classes):
      det_cov = np.linalg.det(self.class_covariances[c, :, :])
      log_prior = np.log(self.prior[c])
      mean_diff = feature_vector - self.class_means[c, :]
      inv_cov = np.linalg.inv(self.class_covariances[c, :, :])
      gauss_kernel = -1/2*np.dot(mean_diff, (np.dot(inv_cov, mean_diff)))

      disc_values[c] = log_prior - 1/2*np.log(det_cov) + gauss_kernel
      log_posteriors[c] = (log_prior -
                           self.num_classes/2*np.log(2*np.pi) -
                           1/2*np.log(det_cov) +
                           gauss_kernel)

    posteriors = np.exp(log_posteriors)
    posteriors = posteriors / np.sum(posteriors) # The posterior should sum to one

    # Check for nonsensical values (remember that the posterior is only used
    # for display, the classification is determined by the disc_values)
    for c in range(self.num_classes):
      if math.isnan(posteriors[c]):
        posteriors[c] = 0
      if posteriors[c] < 1e-12:
        posteriors[c] = 0

    return disc_values, posteriors

  def evaluate(self, test_mask=None, prop_classified_image=None):
    """Compute confusion matrix"""
    if test_mask is not None:
      self.test_mask = test_mask
    if prop_classified_image is not None:
      self.prop_classified_image = prop_classified_image

    self.confusion_matrix = np.zeros([self.num_classes, self.num_classes])
    for i in range(self.prop_classified_image.shape[0]):
      for j in range(self.prop_classified_image.shape[1]):
        if test_mask[i, j] == 0:
          continue
        reference_class = test_mask[i, j]
        proposal_class = self.prop_classified_image[i, j]
        reference_idx = list(self.classes).index(int(reference_class))
        proposal_idx = list(self.classes).index(int(proposal_class))
        self.confusion_matrix[reference_idx, proposal_idx] += 1

  def print_single_measure(self, name, integer=False):
    """Print single measure on the form defined below"""
    print("{0:>5}".format(name), end=' ')
    for c in range(self.num_classes - 1):
      if integer:
        print("{0:>12,.0f}".format(self.eval_measures[name][c]), end=' ')
      else:
        print("{0:>12,.2f}".format(self.eval_measures[name][c]), end=' ')
    if integer:
      print("{0:>12,.0f}".format(self.eval_measures[name][-1]))
    else:
      print("{0:>12,.2f}".format(self.eval_measures[name][-1]))

  def print_evaluation_measures(self, conf_matrix=None):
    """Print info about confusion matrix"""
    if conf_matrix is None:
        cm = self.confusion_matrix
    else:
        cm = conf_matrix

    print('\n')
    print("******************************")
    print("* INFO FROM CONFUSION MATRIX *")
    print("******************************")
    print('\n')

    # Estimated class labels in columns and true class labels in rows. So that
    # cm[i,j] is interpreted as the number of pixels that was labeled with class
    # j from the classifier, when the true pixel label was class i
    print("Confusion matrix")
    for i in range(cm.shape[0]):
      for j in range(cm.shape[1]):
         print('{0:>8,.0f}'.format(cm[i, j]), end=' ')
      print('\n')

    num_ref_pixels_in_class = np.zeros(self.num_classes)
    num_prop_pixels_in_class = np.zeros(self.num_classes)
    for c in range(self.num_classes):
        num_ref_pixels_in_class[c] = sum(cm[c, :])
    for c in range(self.num_classes):
        num_prop_pixels_in_class[c] = sum(cm[:, c])

    total_num_pixels = sum(num_ref_pixels_in_class)
    if total_num_pixels != sum(num_prop_pixels_in_class):
        print("Ugler i mosen i confusion matrixen")

    # True positives
    TP = np.zeros(self.num_classes)
    for c in range(self.num_classes):
      TP[c] = cm[c, c]
    # False positives
    FP = np.zeros(self.num_classes)
    for c in range(self.num_classes):
      FP[c] = sum(cm[:, c]) - TP[c]
    # False negatives
    FN = np.zeros(self.num_classes)
    for c in range(self.num_classes):
      FN[c] = sum(cm[c, :]) - TP[c]
    # True negatives
    TN = np.zeros(self.num_classes)
    for c in range(self.num_classes):
      TN[c] = total_num_pixels - TP[c] - FN[c] - FP[c]
    # Reference positives
    RP = np.zeros(self.num_classes)
    for c in range(self.num_classes):
      RP[c] = TP[c] + FN[c]
    # Reference negatives
    RN = np.zeros(self.num_classes)
    for c in range(self.num_classes):
      RN[c] = TN[c] + FP[c]
    # Proposal positives
    PP = np.zeros(self.num_classes)
    for c in range(self.num_classes):
      PP[c] = TP[c] + FP[c]
    # Proposal negatives
    PN = np.zeros(self.num_classes)
    for c in range(self.num_classes):
      PN[c] = TN[c] + FN[c]

    # Some compound evaluation measures
    tpr = np.zeros(self.num_classes) # Sensitivity, or true positive rate
    tnr = np.zeros(self.num_classes) # Specificity, or true negative rate
    ppv = np.zeros(self.num_classes) # Precision, or positive predictive value
    npv = np.zeros(self.num_classes) # Negative predictive value
    acc = np.zeros(self.num_classes) # Rand accuracy
    iou = np.zeros(self.num_classes) # Jaccard index, intersection over union
    dsc = np.zeros(self.num_classes) # Dice similarity coefficient
    auc = np.zeros(self.num_classes) # Area under ROC curve
    for c in range(self.num_classes):
      tpr[c] = TP[c] / RP[c]
      tnr[c] = TN[c] / RN[c]
      ppv[c] = TP[c] / PP[c]
      npv[c] = TN[c] / PN[c]
      acc[c] = (TP[c] + TN[c]) / total_num_pixels
      iou[c] = TP[c] / (TP[c] + FP[c] + FN[c])
      dsc[c] = 2*TP[c] / (RP[c] + PP[c])
      auc[c] = (tpr[c] + tnr[c]) / 2

    self.eval_measures['RP'] = RP
    self.eval_measures['RN'] = RN
    self.eval_measures['PP'] = PP
    self.eval_measures['PN'] = PN
    self.eval_measures['TP'] = TP
    self.eval_measures['FP'] = FP
    self.eval_measures['FN'] = FN
    self.eval_measures['TN'] = TN
    self.eval_measures['tpr'] = tpr
    self.eval_measures['tnr'] = tnr
    self.eval_measures['ppv'] = ppv
    self.eval_measures['npv'] = npv
    self.eval_measures['acc'] = acc
    self.eval_measures['iou'] = iou
    self.eval_measures['dsc'] = dsc
    self.eval_measures['auc'] = auc

    print(" ------------------- ")
    print(" Evaluation measures ")
    print(" ------------------- ")
    print("\n")

    print("     ", end=' ')
    for c in range(self.num_classes - 1):
        print("    Class {0:>2}".format(self.classes[c]), end=' ')
    print("    Class {0:>2}".format(self.classes[-1]))

    print("=====", end='=')
    for c in range(self.num_classes - 1):
        print("============", end='=')
    print("============")

    self.print_single_measure('RP', integer=True)
    self.print_single_measure('RN', integer=True)
    self.print_single_measure('PP', integer=True)
    self.print_single_measure('PN', integer=True)

    print("-----", end='-')
    for c in range(self.num_classes - 1):
        print("------------", end='-')
    print("------------")

    self.print_single_measure('TP', integer=True)
    self.print_single_measure('FP', integer=True)
    self.print_single_measure('FN', integer=True)
    self.print_single_measure('TN', integer=True)

    print("-----", end='-')
    for c in range(self.num_classes - 1):
        print("------------", end='-')
    print("------------")

    self.print_single_measure('tpr')
    self.print_single_measure('tnr')
    self.print_single_measure('ppv')
    self.print_single_measure('npv')
    self.print_single_measure('acc')
    self.print_single_measure('iou')
    self.print_single_measure('dsc')
    self.print_single_measure('auc')

    print("=====", end='=')
    for c in range(self.num_classes - 1):
        print("============", end='=')
    print("============")
